type 'a stream = S of (unit -> 'a * 'a stream) ;;

let const (x:'a) : 'a stream = 
  let rec loop () = (x, S loop) in 
  S loop
;;

let next ((S f):'a stream) : 'a * 'a stream = f () ;;

let cons (x:'a) (f:unit -> 'a stream) : 'a stream =
  S (fun () -> (x, f()))
;;

let rec take (n:int) (s:'a stream) : 'a list =
  let rec aux n s result =
    if n <= 0 then 
      result
    else
      let (x,s') = next s in
      aux (n-1) s' (x::result)
  in
  List.rev (aux n s [])
;;

let filter (pred:'a -> bool) (s:'a stream) : 'a stream =
  let rec loop (S f) : 'a * 'a stream =
    let (v,s') = f () in
    if pred v then
      (v, S (fun () -> loop s'))
    else
      loop s'
  in
  S (fun () -> loop s)
;;

let map (f:'a -> 'b) (s:'a stream) : 'b stream = 
  let rec loop ((S g):'a stream) : 'b * 'b stream =
    let (v,s') = g () in
    (f v, S (fun () -> loop s'))
  in
  S (fun () -> loop s)
;;

let scan (f:'b -> 'a -> 'b) (base:'b) (s:'a stream) : 'b stream = 
  let rec loop ((S g):'a stream) (base:'b) : 'b * 'b stream =
    let (v,s') = g () in
    let v' = f base v in
    (v', S (fun () -> loop s' v'))
  in
  S (fun () -> loop s base)
;;
